from qgis.core import (
    QgsProcessing,
    QgsProcessingMultiStepFeedback,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterNumber,
    QgsProcessingParameterVectorLayer,
    QgsFeatureRequest,
    QgsFeatureSink,
)
import processing

from PreCourlis.processing.precourlis_algorithm import PreCourlisAlgorithm
from PreCourlis.core.interpolation_area import (
    InterpolationAreaManager,
)


class InterpolateProfilesPerAreasAlgorithm(PreCourlisAlgorithm):
    PROFILES = "PROFILES"
    REACH_AXIS = "REACH_AXIS"
    LATERAL_STEP = "LATERAL_STEP"
    OUTPUT = "OUTPUT"

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterVectorLayer(
                self.PROFILES,
                self.tr("Profiles"),
                types=[QgsProcessing.TypeVectorLine],
                defaultValue=None,
            )
        )
        self.addParameter(
            QgsProcessingParameterVectorLayer(
                self.REACH_AXIS,
                self.tr("Reach axis"),
                types=[QgsProcessing.TypeVectorLine],
                defaultValue=None,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.LATERAL_STEP,
                self.tr("Lateral space step (in m)"),
                defaultValue=None,
            )
        )
        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Interpolated profiles"),
            )
        )

    def checkParameterValues(self, parameters, context):
        ok, msg = super().checkParameterValues(parameters, context)

        # todo

        return ok, msg

    def processAlgorithm(self, parameters, context, model_feedback):
        # Get parameters values
        profiles_layer = self.parameterAsVectorLayer(parameters, self.PROFILES, context)

        reach_axis = self.parameterAsVectorLayer(parameters, self.REACH_AXIS, context)

        lateral_space_step_in_meter = self.parameterAsDouble(
            parameters, self.LATERAL_STEP, context
        )

        (feature_sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            profiles_layer.fields(),
            profiles_layer.wkbType(),
            profiles_layer.sourceCrs(),
        )

        # Get back interpolation areas from profiles layer custom properties
        interpolation_areas = InterpolationAreaManager.readInterpolationAreasFromLayer(
            profiles_layer,
        )

        # Use a multi-step feedback, so that individual child algorithm progress reports
        # are adjusted for the overall progress through the model
        NUMBER_OF_IA = len(interpolation_areas)
        feedback = QgsProcessingMultiStepFeedback(NUMBER_OF_IA, model_feedback)
        outputs = {}

        ia = 0
        sec_id = 1

        # Check that interpolation_areas list are in the right order
        # TODO

        # Processing step :
        for step, interpolation_area in enumerate(interpolation_areas):
            # Check for cancel event
            feedback.setCurrentStep(step)
            if feedback.isCanceled():
                return {}

            # Const variables
            FROM_PROFIL = interpolation_area.from_profil_id
            TO_PROFIL = interpolation_area.to_profil_id
            LONGITUDINAL_STEP = interpolation_area.longitudinal_step

            # Extract by expression from FROM_PROFIL sec_id to TO_PROFIL sec_id
            alg_params = {
                "EXPRESSION": f"sec_id >= {FROM_PROFIL} and sec_id <= {TO_PROFIL}",
                "INPUT": parameters[self.PROFILES],
                "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
            }

            EXTRACT_BY_EXPRESSION_NAME = f"ExtractFrom{FROM_PROFIL}To{TO_PROFIL}"
            outputs[EXTRACT_BY_EXPRESSION_NAME] = processing.run(
                "native:extractbyexpression",
                alg_params,
                context=context,
                feedback=feedback,
                is_child_algorithm=True,
            )

            # Interpolate lines
            INTERPOLATE_LINES_NAME = f"InterpolateLinesFrom{FROM_PROFIL}To{TO_PROFIL}"
            alg_params = {
                "SECTIONS": outputs[EXTRACT_BY_EXPRESSION_NAME]["OUTPUT"],
                "AXIS": reach_axis,
                "CONSTRAINT_LINES": None,
                "LAT_STEP": lateral_space_step_in_meter,
                "LONG_STEP": LONGITUDINAL_STEP,
                "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
                "ATTR_CROSS_SECTION": "sec_id",
            }

            outputs[INTERPOLATE_LINES_NAME] = processing.run(
                "precourlis:interpolate_lines",
                alg_params,
                context=context,
                feedback=feedback,
                is_child_algorithm=True,
            )

            interpolated_lines_layer = context.temporaryLayerStore().mapLayer(
                outputs[INTERPOLATE_LINES_NAME]["OUTPUT"]
            )

            # remove duplicates (but keep master profiles) and update sec_id
            orderby_sec_id_request = QgsFeatureRequest().addOrderBy(
                '"sec_id"', ascending=True
            )
            features_orderby_sec_id = interpolated_lines_layer.getFeatures(
                orderby_sec_id_request
            )
            NUMBER_OF_FEATURES = interpolated_lines_layer.featureCount()

            for i, feature in enumerate(features_orderby_sec_id):
                if i < (NUMBER_OF_FEATURES - 1) or ia == (NUMBER_OF_IA - 1):
                    # add each profiles to sink except the last one
                    # but we still add the last one of the last interpolation area
                    feature["sec_id"] = sec_id
                    sec_id += 1

                    feature_sink.addFeature(
                        feature,
                        QgsFeatureSink.FastInsert,
                    )

            ia += 1

        # Keep "master" profiles (OK) and show this information (how ?) TODO

        # Return output layer id
        return {self.OUTPUT: dest_id}

    def name(self):
        return "interpolate_profiles_per_areas"

    def displayName(self):
        return self.tr("Interpolate profiles per areas")

    def group(self):
        return self.tr("Interpolate")

    def groupId(self):
        return "Interpolate"

    def createInstance(self):
        return InterpolateProfilesPerAreasAlgorithm()
